//#define DEBUG
#include "gpsntp.h"

#if getenv("PIN:A0")
#use fast_io(A)
#endif
#if getenv("PIN:B0")
#use fast_io(B)
#endif
#if getenv("PIN:C0")
#use fast_io(C)
#endif
#if getenv("PIN:D0")
#use fast_io(D)
#endif
#if getenv("PIN:E0")
#use fast_io(E)
#endif

#include "queue.c"

int8 hexchar(char c)
{
  if (c >= '0' && c <= '9') return c - '0';
  if (c >= 'A' && c <= 'F') return c - 'A' + 10;
  return 0;
}

signed long a_to_l(char *s)
{
  signed long r;
  int1 sign;
  r = 0;
  sign = (*s == '-');
  if (sign) s++;
  while (*s >= '0' && *s <= '9') r = r * 10 + (*s++ - '0');
  if (sign) r = -r;
  return r;
}

int1 status_ok = 0;

/* * * * * * * * * *
 * External EEPROM *
 * * * * * * * * * */

#define PIN_EE1_CS PIN_D0
#define PIN_EE2_CS PIN_D1
#define PIN_EE3_CS PIN_D2
#define PIN_EE4_CS PIN_D3

#define PIN_DAC_CS    PIN_EE1_CS
#define PIN_LEDDSP_CS PIN_EE2_CS
#define PIN_MAC_EE_CS PIN_EE4_CS

/* * * * * * * *
 * Status LEDs *
 * * * * * * * */

#define LED_1A PIN_D6
#define LED_1K PIN_C2
#define LED_2A PIN_D5
#define LED_2K PIN_D4
#define LED_OFF 0
#define LED_RED 1
#define LED_GREEN 2

void led_init(void)
{
  output_drive(LED_1A);
  output_drive(LED_1K);
  output_drive(LED_2A);
  output_drive(LED_2K);
}

void led_set(int8 led, int8 state)
{
  int1 a, k;
  a = k = 0;
  if (state == LED_RED  ) a = 1;
  if (state == LED_GREEN) k = 1;
  if (led == 1)
  {
    output_bit(LED_1A, a);
    output_bit(LED_1K, k);
  }
  if (led == 2)
  {
    output_bit(LED_2A, a);
    output_bit(LED_2K, k);
  }
}

/* * * * * * * * * * * * * * *
 * External Real-Time Clock  *
 * * * * * * * * * * * * * * */

#define RTC_USE_SPI
#define RTC_CE   PIN_D7
#define RTC_BEGIN setup_spi(SPI_MASTER|SPI_L_TO_H|                SPI_CLK_DIV_16)
#define RTC_END   setup_spi(SPI_MASTER|SPI_L_TO_H|SPI_XMIT_L_TO_H|SPI_CLK_DIV_4 )
#include <ds1305.c>
int8 RtcGetCentury(void)
{
  return rtc_read_nvr(0);
}
void RtcSetCentury(int8 c)
{
  rtc_write_nvr(0, c);
}
signed int8 RtcGetTimezone(void)
{
  return rtc_read_nvr(1);
}
void RtcSetTimezone(signed int8 tz)
{
  rtc_write_nvr(1, tz);
}

int1 isLeapYear(int16 cyr)
{
  return (cyr % 4 == 0) - (cyr % 100 == 0) + (cyr % 400 == 0);
}

// calculate day of week (0=Sunday,6=Saturday), from cyr (eg 2007), mo (1-12), dy (1-31)
// http://en.wikipedia.org/wiki/Talk:Calculating_the_day_of_the_week
int8 dayOfWeek(int16 cyr, int8 mo, int8 dy)
{
  int8 f;
  int16 y;
  f = (14 - mo) / 12;
  y = cyr - f;
  return ((int16)dy + y + 31 * (int16)(mo + 12 * f - 2) / 12 + y / 4 - y / 100 + y / 400) % 7;
}

// largest day of month
int8 maxDom(int16 cyr, int8 mo)
{
  const int8 maxDaysInMonth[]={31,28,31,30,31,30,31,31,30,31,30,31};
  if (mo == 2 && isLeapYear(cyr)) return 29;
  return maxDaysInMonth[mo - 1];
}

// returns day of year (1-366)
int16 dayOfYear(int16 cyr, int8 mo, int8 dy)
{
  int16 ret;
  ret = dy;
  for (--mo ; mo > 0; --mo) ret += maxDom(cyr, mo);
  return ret;
}

// add the timezone offset (in 15min units) to the given date and time)
void applyTimezone(signed int16 &cyr, signed int8 &mo, signed int8 &dy,
                   signed int8 &hr, signed int8 &mn, signed int8 &sc,
                   signed int8 tz)
{
  mn += (tz % 4) * 15;
  if (mn <   0) { --hr; mn += 60; }
  if (mn >= 60) { ++hr; mn -= 60; }
  hr += (tz / 4);
  if (hr <   0) { --dy; hr += 24; }
  if (hr >= 24) { ++dy; hr -= 24; }
  if (dy <   1) { --mo; dy = maxDom(cyr, mo); }
  if (dy > maxDom(cyr, mo)) { ++mo; dy = 1; }
  if (mo <   1) { --cyr; mo = 12; }
  if (mo >  12) { ++cyr; mo = 1; }
}

/* * * * * * * * * * *
 * Millisecond timer *
 * * * * * * * * * * */

#include "timesvr.h"

int8  millisovr; // milliseconds overflow count
int16 millis; // milliseconds 0..999
int16 pps_millis[2]; // milliseconds since the most recent two PPS
int16 serial_lastch_millis; // milliseconds since most recently received serial character
int32 uptime = 0; // seconds since reset
int32 millisup = 0; // milliseconds since reset

signed int16 status_led1 = 0;
int16 status_led2 = 0;

#int_CCP1
void CCP1_isr(void)
{
  millisup++;
  if (++millis >= 1000)
  {
    millis = 0;
    ++millisovr;
    ++uptime;
  }
  if (serial_lastch_millis != 0xFFFF) ++serial_lastch_millis;
  if (pps_millis[0] != 0xFFFF) ++pps_millis[0];
  if (pps_millis[1] != 0xFFFF) ++pps_millis[1];
  if (status_ok)
  {
    if (status_led1 > 0) led_set(1, --status_led1 == 0 ? LED_OFF : LED_GREEN);
    if (status_led1 < 0) led_set(1, ++status_led1 == 0 ? LED_OFF : LED_RED  );
    if (status_led2 == 0)
    {
      led_set(2, TimeServerHasSync ? LED_GREEN : LED_RED);
    }
    else
    {
      led_set(2, LED_OFF);
      --status_led2;
    }
  }
  else
  {
    // problem status - toggle LEDs like railway crossing lights
    static int1 on = 0;
    if (on == (millis >= 500))
    {
      on = (millis < 500);
      led_set(1, on ? LED_RED : LED_OFF);
      led_set(2, on ? LED_OFF : LED_RED);
    }
  }
}

// obtain a 1ms interrupt tick by setting timer 3 as fast as possible, then
// using CCP1 to interrupt and reset so no ticks are lost handling the interrupt
void init_millis(void)
{
  CCP_1 = (getenv("CLOCK")/4/1000-1);
  set_timer3(0);
  setup_timer_3(T3_INTERNAL | T3_DIV_BY_1 | T3_CCP1_TO_2);
  setup_ccp1(CCP_COMPARE_RESET_TIMER);
  enable_interrupts(INT_CCP1);
}

// seconds up
int32 get_uptime(void)
{
  int32 ret;
  disable_interrupts(GLOBAL);
  ret = uptime;
  enable_interrupts(GLOBAL);
  return ret;
}

// milliseconds up
int32 get_millisup(void)
{
  int32 ret;
  disable_interrupts(GLOBAL);
  ret = millisup;
  enable_interrupts(GLOBAL);
  return ret;
}

/* * * * * * * * * * * * * * *
 * Internal Real Time Clock  *
 * * * * * * * * * * * * * * */

// An internal clock is needed to ensure the regular clock (year, month, day,
// hour, minute, second) is in sync with the milliseconds.

struct {
  int8  sc;  // seconds 0..59
  int8  mn;  // minutes 0..59
  int8  hr;  // hour 0..23
  int8  dy;  // day of month 1..31
  int8  mo;  // month 1..12
  int16 cyr; // year with century
} intrtc;

// Initialise the internal RTC from the external RTC
void intrtc_init(void)
{
  int8 ct, yr, dow;
  ct = RtcGetCentury();
  rtc_get_date(intrtc.dy, intrtc.mo, yr, dow);
  intrtc.cyr = (int16)ct * 100 + yr;
  if (intrtc.mo < 1 || intrtc.mo > 12) intrtc.mo = 1;
  if (intrtc.dy < 1 || intrtc.dy > maxDom(intrtc.cyr, intrtc.mo)) intrtc.dy = 1;
  rtc_get_time(intrtc.hr, intrtc.mn, intrtc.sc);
  if (intrtc.hr > 23) intrtc.hr = 0;
  if (intrtc.mn > 59) intrtc.mn = 0;
  if (intrtc.sc > 59) intrtc.sc = 0;
  millis = 0;
  millisovr = 0;
}

// update RTC if the milliseconds have overflowed
// returns the current millisecond
int16 intrtc_update(void)
{
  int16 ret;
  int8 ovr;
  int1 update_dsp;
  disable_interrupts(GLOBAL);
  ret = millis;
  ovr = millisovr;
  update_dsp = FALSE;
  millisovr = 0;
  enable_interrupts(GLOBAL);
  while (ovr--)
  {
    update_dsp = TRUE;
    if (++intrtc.sc >= 60)
    {
      intrtc.sc = 0;
      if (++intrtc.mn >= 60)
      {
        intrtc.mn = 0;
        if (++intrtc.hr >= 24)
        {
          intrtc.hr = 0;
          if (++intrtc.dy > maxDom(intrtc.cyr, intrtc.mo))
          {
            intrtc.dy = 1;
            if (++intrtc.mo > 12)
            {
              intrtc.mo = 1;
              intrtc.cyr++;
            }
          }
        }
      }
    }
  }
  if (update_dsp)
  {
    ovr = intrtc.hr + 24 + RtcGetTimezone() / 4;
    while (ovr >= 24) ovr -= 24;
    RTC_BEGIN;
    output_low(PIN_LEDDSP_CS);
    spi_write(ovr | 0x80);
    output_high(PIN_LEDDSP_CS);
    delay_us(80);
    output_low(PIN_LEDDSP_CS);
    spi_write(intrtc.mn);
    output_high(PIN_LEDDSP_CS);
    RTC_END;
    rtc_set_datetime(intrtc.dy, intrtc.mo, intrtc.cyr % 100, dayOfWeek(intrtc.cyr, intrtc.mo, intrtc.dy), intrtc.hr, intrtc.mn, intrtc.sc);
    RtcSetCentury(intrtc.cyr / 100);
  }
  return ret;
}

// Get the date/time from the internal RTC
void get_datetime(int16 &cyr, int8 &mo, int8 &dy, int8 &hr, int8 &mn, int8 &sc, int16 &m)
{
  m = intrtc_update();
  sc  = intrtc.sc;
  mn  = intrtc.mn;
  hr  = intrtc.hr;
  dy  = intrtc.dy;
  mo  = intrtc.mo;
  cyr = intrtc.cyr;
}

// Set the date/time of the internal RTC, backs up to external RTC
void set_datetime(int16 cyr, int8 mo, int8 dy, int8 hr, int8 mn, int8 sc, int16 m)
{
  int8 o;
  for (o = 0; m >= 1000; o++) m -= 1000;
  disable_interrupts(GLOBAL);
  millisovr = o;
  millis = m;
  enable_interrupts(GLOBAL);
  intrtc.sc  = sc;
  intrtc.mn  = mn;
  intrtc.hr  = hr;
  intrtc.dy  = dy;
  intrtc.mo  = mo;
  intrtc.cyr = cyr;
  intrtc_update();
}

/* * * * * * * * * *
 * Expansion Port  *
 * * * * * * * * * */

#define PIN_EXP1 PIN_A0
#define PIN_EXP2 PIN_A1
#define PIN_EXP3 PIN_A2
#define PIN_EXP4 PIN_A3
#define PIN_EXP5 PIN_A4
#define PIN_EXP6 PIN_A5
#define PIN_EXP7 PIN_E0
#define PIN_EXP8 PIN_E1

void exp_init(void)
{
  output_low(PIN_EXP1);
  output_low(PIN_EXP2);
  output_low(PIN_EXP3);
  output_low(PIN_EXP4);
  output_low(PIN_EXP5);
  output_low(PIN_EXP6);
  output_low(PIN_EXP7);
  output_low(PIN_EXP8);
  output_drive(PIN_EXP1);
  output_drive(PIN_EXP2);
  output_drive(PIN_EXP3);
  output_drive(PIN_EXP4);
  output_drive(PIN_EXP5);
  output_drive(PIN_EXP6);
  output_drive(PIN_EXP7);
  output_drive(PIN_EXP8);
}

/* * * * * * * * * *
 * Internal EEPROM *
 * * * * * * * * * */

// MAC Address Notes:
// First byte (02) : bit 1 must be set to indicate a "locally administered"
// address, bit 0 must be clear for an "inidividual" rather than a "multicast"
// address. Bits 7..2 can be anything.
#rom int8 0xF00000 = {
  192,168,1,4,                                   // IP address
  255,255,255,0,                                 // Netmask
  192,168,1,1,                                   // Gateway
  192,168,1,1,                                   // DNS server
  't','i','m','e',0,0,0,0,0,0,0,0,0,0,0,0,       // NetBIOS name
  0,                                             // use DHCP (0=no,1=yes)
  44,1,                                          // Timeserver MaxDiff (300)
  0,0,                                           // Timeserver Broadcast Interval (0)
  0,0, // padding
  'a','d','m','i','n',':','t','i','m','e',0,0,0,0,0,0,0,0,0,0 // web login
}
/*******************************************************************
 * IMPORTANT: Initializers above must match structure layout below *
 *******************************************************************/

#define LOGIN_SIZE 20
struct {
  int32 MyIPAddr;                     // ip address
  int32 MyMask;                       // net mask
  int32 MyGateway;                    // gateway
  int32 PrimaryDNSServer;             // DNS server
  char NetBIOSName[16];               // NetBIOS name
  int8 UseDHCP;                       // use DHCP: 0 = no, 1 = yes
  int16 TimeServerGPS_MaxDiff;        // server must already be within 5min of correct time in order to sync
  int16 TimeServer_BroadcastInterval; // interval in seconds between broadcasts (zero disables)
  int16 unused1;
  char Login[LOGIN_SIZE];
} Config;

#define CONFIG_EEPROM_ADDR 0
void ConfigLoad(void)
{
  int8 i;
  int8 *p;
  for (i = 0, p = &Config; i < sizeof(Config); i++, p++) *p = read_eeprom(CONFIG_EEPROM_ADDR + i);
}

void ConfigSave(void)
{
  int8 i;
  int8 *p;
  for (i = 0, p = &Config; i < sizeof(Config); i++, p++) write_eeprom(CONFIG_EEPROM_ADDR + i, *p);
}

/* * * * * * * * * *
 * PPS Processing  *
 * * * * * * * * * */

// VCXO is 100ppm and pullable +/-100ppm by 0-3.3V.
// (3300/2)/(100/1000000)=16,500,000
// i.e. the smallest 1mV change should result in a 1 in 16,500,000 change.
// At 25,000,000MHz, Timer 1 counts at 6,250,000MHz. 16,500,000/6,250,000=2.64,
// round up to 3.
#define PPS_COUNT 3
#define PPS_TICKS (getenv("CLOCK")/4*PPS_COUNT)
#define PPS_DAC_MAX 3300

extern int1 gps_valid;

int8 pps_state;
int1 pps_update_dac;
int16 pps_dac_value;
int16 pps_dac_adjust;

int16 timer1_ovr;

#int_TIMER1
void TIMER1_isr(void)
{
  timer1_ovr++;
}

int8 pps_cnt;

void pps_init(void)
{
  timer1_ovr = (PPS_TICKS >> 16) + 1;
  pps_state = PPS_COUNT;
  pps_update_dac = TRUE;
  pps_dac_value = PPS_DAC_MAX / 2;
  pps_dac_adjust = pps_dac_value;
  pps_cnt = 0;
  pps_millis[0] = 0;
  pps_millis[1] = 0;
  enable_interrupts(INT_TIMER1);
  ext_int_edge(1, H_TO_L);
  enable_interrupts(INT_EXT1);
}

#int_EXT1
void EXT1_isr(void)
{
  // falling edge on the pin = rising edge from the GPS (MAX3232 inverts the logic)
  
  // TESTING
  if (!gps_valid && TimeServerHasSync)
  {
    // only PPS during position locks are any good - make any timing invalid
    timer1_ovr = (PPS_TICKS >> 16) + 1;
    pps_state = PPS_COUNT;
  }
  
  if (pps_state == PPS_COUNT)
  {
    int32 v;
    v = make32(timer1_ovr, get_timer1());
    set_timer1(0);
    timer1_ovr = 0;
    // ideally, v should now equal PPS_TICKS
    // make sure v is within 0.02% of correct
    // the VCXO should be within 100ppm (0.01%)
    if ((v > (PPS_TICKS-PPS_TICKS/5000)) && (v < (PPS_TICKS+PPS_TICKS/5000)))
    {
      // adjust the VCXO proportionally to how far it's out
      // make sure the adjustment doesn't go out of range (0 to 3.3V)
      if (v < PPS_TICKS)
      {
        pps_dac_adjust = (int16)(PPS_TICKS - v);
        pps_dac_adjust += 6;
        pps_dac_adjust /= 8;
        // adjust and range-check DAC
        pps_dac_value += pps_dac_adjust;
        if (pps_dac_value > PPS_DAC_MAX) pps_dac_value = PPS_DAC_MAX;
        // output new DAC value
        pps_update_dac = TRUE;
      }
      if (v > PPS_TICKS)
      {
        pps_dac_adjust = (int16)(v - PPS_TICKS);
        pps_dac_adjust += 6;
        pps_dac_adjust /= 8;
        // adjust and range-check DAC
        if (pps_dac_adjust > pps_dac_value) pps_dac_value = 0;
        else pps_dac_value -= pps_dac_adjust;
        // output new DAC value
        pps_update_dac = TRUE;
      }
    }
    pps_state = 0;
  }
  pps_state++;
  if (serial_lastch_millis < 10)
  {
    // still processing serial data
    if (pps_cnt == 2)
    {
      // too many counts - discard oldest
      pps_millis[0] = pps_millis[1];
    }
    else
    {
      if (pps_cnt == 0) pps_millis[0] = 0;
      pps_cnt++;
    }
  }
  else
  {
    // not processing serial data - new pulse
    pps_cnt = 1;
    pps_millis[0] = 0;
  }
  pps_millis[1] = 0;
}

// This must be called once when an RMC message is received.
int1 pps_adjust_millis(int16 &m)
{
  int16 adj;
  adj = 0;
  disable_interrupts(GLOBAL);
  if (pps_cnt != 0)
  {
    adj = pps_millis[0];
    pps_millis[0] = pps_millis[1];
    pps_cnt--;
    if (adj >= 1500 && pps_cnt != 0)
    {
      adj = pps_millis[0];
      pps_cnt = 0;
    }
  }
  enable_interrupts(GLOBAL);
  if (adj > 0 && adj < 1500)
  {
    m += adj;
    return TRUE;
  }
  return FALSE;
}

/* * * * * * * *
 * Serial Port *
 * * * * * * * */

#define PIN_DTRRTS PIN_C1
#define PIN_DSRCTS PIN_C0
#define PIN_OUTLVL PIN_B4
#define PIN_SELOUT PIN_B3
#define SERIAL_INPUT_BUFFER_SIZE 1000
#define SERIAL_OUTPUT_BUFFER_SIZE 100
char serial_input_buffer[SERIAL_INPUT_BUFFER_SIZE];
char serial_output_buffer[SERIAL_OUTPUT_BUFFER_SIZE];
int16 serial_input_buffer_head;
int16 serial_input_buffer_tail;
int16 serial_output_buffer_head;
int16 serial_output_buffer_tail;
int8 serial_baud;

// Baud rate
// Decimal Hex     Sum-of-bytes
//   1200 0x0004B0 0xB4
//   2400 0x000960 0x69
//   4800 0x0012C0 0xD2
//   9600 0x002580 0xA5
//  19200 0x004B00 0x4B
//  38400 0x009600 0x96
//  57600 0x00E100 0xE1
// 115200 0x01C200 0xC3
void serial_setup_uart(int32 v)
{
  int8 t;
  t = make8(v, 2) + make8(v, 1) + make8(v, 0);
  if (t == 0 || t == 0xB4) { setup_uart(  1200); t = 0; }
  if (t == 1 || t == 0x69) { setup_uart(  2400); t = 1; }
  if (t == 2 || t == 0xD2) { setup_uart(  4800); t = 2; }
  if (t == 3 || t == 0xA5) { setup_uart(  9600); t = 3; }
  if (t == 4 || t == 0x4B) { setup_uart( 19200); t = 4; }
  if (t == 5 || t == 0x96) { setup_uart( 38400); t = 5; }
  if (t == 6 || t == 0xE1) { setup_uart( 57600); t = 6; }
  if (t == 7 || t == 0xC3) { setup_uart(115200); t = 7; }
  if (t < 8) serial_baud = t;
}

void serial_init(void)
{
  serial_lastch_millis = 0;
  serial_input_buffer_head = 0;
  serial_input_buffer_tail = 0;
  serial_output_buffer_head = 0;
  serial_output_buffer_tail = 0;
  serial_setup_uart(4800);
  
  // setup RS232 signals
  output_low(PIN_DTRRTS);
  output_drive(PIN_DTRRTS);
  output_float(PIN_DSRCTS);
  output_low(PIN_OUTLVL);
  output_drive(PIN_OUTLVL);
  output_low(PIN_SELOUT);
  output_drive(PIN_SELOUT);
  
  enable_interrupts(INT_RDA);
}

#int_RDA
void RDA_isr(void) 
{
  serial_lastch_millis = 0;
  serial_input_buffer[serial_input_buffer_tail] = fgetc(SERIAL);
  if (++serial_input_buffer_tail >= SERIAL_INPUT_BUFFER_SIZE)
    serial_input_buffer_tail = 0;
  if (serial_input_buffer_tail == serial_input_buffer_head)
    if (++serial_input_buffer_head >= SERIAL_INPUT_BUFFER_SIZE)
      serial_input_buffer_head = 0;
}

#int_TBE
void TBE_isr(void) 
{
  fputc(serial_output_buffer[serial_output_buffer_head], SERIAL);
  if (++serial_output_buffer_head >= SERIAL_OUTPUT_BUFFER_SIZE)
    serial_output_buffer_head = 0;
  if (serial_output_buffer_head == serial_output_buffer_tail)
    disable_interrupts(INT_TBE);
}

// add a character to the outgoing (PIC->serial port) queue
void bputc(int8 c)
{
  disable_interrupts(GLOBAL);
  serial_output_buffer[serial_output_buffer_tail] = c;
  if (serial_output_buffer_head == serial_output_buffer_tail) enable_interrupts(INT_TBE);
  if (++serial_output_buffer_tail >= SERIAL_OUTPUT_BUFFER_SIZE)
    serial_output_buffer_tail = 0;
  if (serial_output_buffer_tail == serial_output_buffer_head)
    if (++serial_output_buffer_head >= SERIAL_OUTPUT_BUFFER_SIZE)
      serial_output_buffer_head = 0;
  enable_interrupts(GLOBAL);
}

int1 bkbhit(void)
{
  int1 ret;
  disable_interrupts(GLOBAL);
  ret = (serial_input_buffer_head != serial_input_buffer_tail);
  enable_interrupts(GLOBAL);
  return ret;
}

int8 bgetc(void)
{
  int8 c;
  disable_interrupts(GLOBAL);
  c = serial_input_buffer[serial_input_buffer_head];
  if (++serial_input_buffer_head >= SERIAL_INPUT_BUFFER_SIZE)
    serial_input_buffer_head = 0;
  enable_interrupts(GLOBAL);
  return c;
}

/* * * * * * * * * * * * * * * * * *
 * TCP/IP / Ethernet Configuration *
 * * * * * * * * * * * * * * * * * */

#define TELNETQ_TX_SIZE 250
/*
#define TELNET_GET_BAUD serial_get_baud()
#define TELNET_SET_BAUD serial_setup_uart
#define TELNET_GET_DTR ((!(Config.SerialConfig & 0x80)) && input_state(PIN_DTRRTS))
#define TELNET_SET_DTR(STATE) if (!(Config.SerialConfig & 0x80)) output_bit(PIN_DTRRTS, STATE)
#define TELNET_GET_RTS ((Config.SerialConfig & 0x80) && input_state(PIN_DTRRTS))
#define TELNET_SET_RTS(STATE) if (Config.SerialConfig & 0x80) output_bit(PIN_DTRRTS, STATE)
*/

#define HTTP_CLOCK_ENABLE
#define HTTP_CLOCK_GET { int16 millis; get_datetime(cyr, mo, dy, hr, mn, sc, millis); }

#define ___TCPIP_STACK_CONFIGURATION
#define TCP_NO_WAIT_FOR_ACK   FALSE
#define MAX_SOCKETS                     5 // HTTP,TELNET,DAYTIME,TIME,NETBIOS (was 7)
#define MAX_UDP_SOCKETS                 7 // DHCP,DAYTIME,TIME,NTP,broadcast NTP/reply UDP,DNS,NBNS
#define MAC_TX_BUFFER_SIZE          1024 //do not modify this line
#define MAC_TX_BUFFER_COUNT         1    //do not modify this line
#define MAC_RX_BUFFER_SIZE              MAC_TX_BUFFER_SIZE  //do not modify this line unless you are certain you know what you're doing

#define PIN_ENC_MAC_CS  PIN_A6 // output
#define PIN_ENC_MAC_RST PIN_E2 // output
#define PIN_ENC_MAC_SO  PIN_C4   // PIC <<<< ENC input
#define PIN_ENC_MAC_SI  PIN_C5   // PIC >>>> ENC output
#define PIN_ENC_MAC_CLK PIN_C3 // output
#define PIN_ENC_MAC_INT PIN_B0 // input
#define PIN_ENC_MAC_WOL PIN_B1 // input
#define ENC_MAC_USE_SPI TRUE      //due to an errata in the ENC28J60, you should always use HW SPI to assure that SPI clock is over 8MHz!
#define mac_enc_spi_tris_init {  \
  output_float(PIN_ENC_MAC_SO ); \
  output_low  (PIN_ENC_MAC_SI ); \
  output_drive(PIN_ENC_MAC_SI ); \
  output_low  (PIN_ENC_MAC_CLK); \
  output_drive(PIN_ENC_MAC_CLK); \
  output_float(PIN_ENC_MAC_INT); \
  output_float(PIN_ENC_MAC_WOL); \
  output_high (PIN_ENC_MAC_CS ); \
  output_drive(PIN_ENC_MAC_CS ); \
  output_high (PIN_ENC_MAC_RST); \
  output_drive(PIN_ENC_MAC_RST); \
}

//#include "telnet.h"

#define STACK_USE_MAC TRUE
#define STACK_USE_MCPENC TRUE
#define STACK_USE_UDP TRUE
#define STACK_USE_DHCP TRUE
#define STACK_USE_ARP TRUE
#define STACK_USE_DNS TRUE
#define STACK_USE_HTTP TRUE
#define HTTP_USE_CONTENT_TYPE TRUE
//#define STACK_USE_SNMP TRUE
#define STACK_USE_TCP TRUE
#define STACK_USE_ICMP TRUE
#define STACK_USE_IP_GLEANING FALSE  //Disable Microchip's propietary auto-IP configuration
#include "tcpip/stacktsk.c"
#ifdef TELNET_PORT
#include "telnet.c"
#endif
#include "timesvr.c"

#if sizeof(AppConfig.NetBIOSName) != 16
#error NetBIOSName size assumed to be 16 (fix EEPROM initializer above)
#endif

// parses a text string into an IP address
// only updates the given address (and returns TRUE) if all the dotted quads are
// 0..255 and they aren't all 0 or 255 (otherwise returns FALSE)
int1 process_ip_val(char *val, IP_ADDR *ip)
{
  IP_ADDR tip;
  char *p;
  int8 i;
  int16 tv;
  tip.Val = 0;
  for (i = 0, tv = 0, p = val; ; p++)
  {
    if (*p == '.' || *p == 0)
    {
      if (i > 3) return FALSE; // out of range
      tip.v[i++] = tv;
      if (*p == 0) break;
      tv = 0;
    }
    else if (*p >= '0' && *p <= '9')
    {
      tv = tv * 10 + (*p - '0');
      if (tv > 255) return FALSE; // out of range
    }
    else
      return FALSE; // invalid character
  }
  if (i != 4 || tip.Val == 0 || tip.Val == 0xFFFFFFFF) return FALSE;
  ip->Val = tip.Val;
  return TRUE;
}

/* * * * * * * *
 * Web Server  *
 * * * * * * * */
 
/*
 * Style Sheet
 */

const char CSS_STYLE_SHEET[]=
"*{font-family:sans-serif}"
"circle{fill:none;stroke:#ccc}"
"path,g g.s circle{stroke:#000}"
"text{text-anchor:middle;font-size:9px}"
"g.n,g.s{fill:#009}"
"body{background:#eef}"
"h1{border-bottom:1px solid #000;margin-top:0}"
"table{border-collapse:collapse}"
"th{text-align:right}"
"th,td{border:2px solid #009;background:#fff;padding:.3em}";

/*
 *  Index Page
 */

const char HTML_INDEX_PAGE[]=
"%D<title>GPS NTP Server</title>%L%S>"
"<h1>GPS NTP Server</h1><ul><li><a href=status>Status</a><li>Configure<ul>"
"<li><a href=tcpip>TCP/IP</a>"
"<li><a href=time>Time</a>"
"<li><a href=server>Time Server</a>"
"<li><a href=login>Login</a>"
"</ul></ul>%F";

/*
 *  Status Page
 */

const char HTML_STATUS_PAGE[]=
"%D<title>Status</title>%L%S>"
"<meta http-equiv=refresh content=30>"
"<h1>Status</h1>"
"<table border=0>"
"<tr>"
"<th>Date/Time:<td>%0"
"<th>Uptime:<td>%1"
"<tr>"
"<th>GPS Location:<td>%2%3"
"<th>VCXO:<td>%4mV"
"</table>"
"<p><object type=\"image/svg+xml\" data=sats width=480 height=240><param name=src value=sats>"
"<a href=sats>View satellites</a></object>%F";

void status_format_char(char id, char *str)
{
  int16 cyr, m;
  signed int8 mo, dy, hr, mn, sc, tz;
  get_datetime(cyr, mo, dy, hr, mn, sc, m);
  // UTC -> localtime
  tz = RtcGetTimezone();
  applyTimezone(cyr, mo, dy, hr, mn, sc, tz);
  switch (id)
  {
  case '0':
    // "1999-12-31 23:59:59 +00:00"
    sprintf(str, "%04lu-%02u-%02u %02u:%02u:%02u %03u:00", cyr, mo, dy, hr, mn, sc, tz / 4);
    str[20] = (tz >= 0) ? '+' : '-';
    if (tz % 4 != 0) str[24] = '3';
    if (!TimeServerHasSync)
    {
      // trailing asterisk indicates no sync
      str[26] = '*';
      str[27] = 0;
    }
    break;
  case '1':
    {
      int32 diff;
      diff = get_uptime();
      sprintf(str, "%lu:%02lu:%02lu", diff / 3600, diff / 60 % 60, diff % 60);
    }
    break;
  case '2':
    
    sprintf(str, "%s&deg;%s'%c %s&deg;%s'%c",
            &TimeServerLatitude[0], &TimeServerLatitude[3], TimeServerLatHemi,
            &TimeServerLongitude[0], &TimeServerLongitude[4], TimeServerLongHemi);
    break;
  case '3':
    str[0] = 0;
    str[1] = 0;
    if (!gps_valid) str[0] = '*';
    break;
  case '4':
    sprintf(str, "%lu", pps_dac_value);
    break;
  }
}

const char SVG_STATUS_PAGE[]=
"<?xml-stylesheet%S?>"
"<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">"
"<svg version=\"1.1\" baseProfile=\"full\" xmlns=\"http://www.w3.org/2000/svg\" width=\"480\" height=\"240\">"
"<title>GPS Satellites</title>"
"<g transform=\"translate(120,120) rotate(-90) scale(1.25)\">"
"<circle r=\"90\"/><circle r=\"45\"/>%0"
"%1%1%1%1%1%1%1%1%1%1%1%1"
"<g transform=\"rotate(90) translate(100,50)\">"
"<path d=\"M0 0h175M0-100h175\"/>"
"%2%2%2%2%2%2%2%2%2%2%2%2"
"</g></g></svg>";

int8 svg_status_sat;
void svg_status_format_char(char id, char *str)
{
  if (id == '0') svg_status_sat = 12;
  else
  {
    if (svg_status_sat == 12) svg_status_sat = 0;
    if (id == '1')
    {
      int16 az;
      az = sat_azimuth[svg_status_sat];
      if (az > 90) az = 450 - az;
      else az = 90 - az;
      sprintf(str, "<g class=\"s\" transform=\"rotate(%lu) translate(%u,0) rotate(%lu)\"><circle r=\"7\"/><text y=\"3\">%u</text></g>",
              sat_azimuth[svg_status_sat], 90 - sat_elevation[svg_status_sat], az, sat_id[svg_status_sat]);
    }
    if (id == '2')
    {
      int8 x;
      x = svg_status_sat * 15;
      sprintf(str, "<g class=\"s\"><text x=\"%u\" y=\"10\">%u</text><rect x=\"%u\" y=\"-%u\" width=\"10\" height=\"%u\"/></g>",
                   x + 5, sat_id[svg_status_sat], x, sat_dbm[svg_status_sat], sat_dbm[svg_status_sat]);
    }
    if (id == '1' || id == '2')
    {
      if (sat_dbm[svg_status_sat] == 0) str[10] = 'n';
    }
    if (sat_id[svg_status_sat] == 0) str[0] = 0;
    ++svg_status_sat;
  }
}

/*
 *  TCP/IP Configuration Page
 */

const char HTML_TCPIP_CONFIG_PAGE[]=
"%D<title>TCP/IP</title>%L%S>"
"<h1>TCP/IP</h1>"
"<p>MAC=%0"
"<p>ENC28J60:Rev %8"
"<form action=\"/tcpip\" method=post>"
"<table border=1>"
"<tr><th rowspan=2 valign=top>Method:<td><input type=radio name=A value=1%1>DHCP"
"<tr><td valign=top><input type=radio name=A value=0%2>Manual"
"<table>"
"<tr><th>IP:<td><input name=B value=\"%3\">"
"<tr><th>Netmask:<td><input name=C value=\"%4\">"
"<tr><th>Gateway:<td><input name=D value=\"%5\">"
"<tr><th>DNS:<td><input name=E value=\"%6\">"
"</table>"
"<tr><th>NetBIOS Name:<td><input name=F value=\"%7\">"
"</table>"
"<p><input type=submit>"
"</form>%F";

void tcpip_config_cgi(char *key, char *val)
{
  int8 i;
  IP_ADDR t;
  switch (key[0])
  {
  case 'A':
    if (DHCPIsDisabled())
    {
      if (val[0] != '0')
      {
        // enable DHCP
        Config.UseDHCP = 1;
        smDHCPState = SM_DHCP_INIT_FIRST_TIME;
        DHCPReset();
      }
    }
    else
    {
      if (val[0] == '0')
      {
        // disable DHCP
        Config.UseDHCP = 0;
        smDHCPState = SM_DHCP_INIT_FIRST_TIME;
        DHCPReset();
        DHCPDisable();
      }
    }
    break;
  case 'B':
    if (process_ip_val(val, &t))
    {
      Config.MyIPAddr = t.Val;
      if (!DHCPIsDisabled()) AppConfig.MyIPAddr.Val = t.Val;
    }
    break;
  case 'C':
    if (process_ip_val(val, &t))
    {
      Config.MyMask = t.Val;
      if (!DHCPIsDisabled()) AppConfig.MyMask.Val = t.Val;
    }
    break;
  case 'D':
    if (process_ip_val(val, &t))
    {
      Config.MyGateway = t.Val;
      if (!DHCPIsDisabled()) AppConfig.MyGateway.Val = t.Val;
    }
    break;
  case 'E':
    if (process_ip_val(val, &t))
    {
      Config.PrimaryDNSServer = t.Val;
      if (!DHCPIsDisabled()) AppConfig.PrimaryDNSServer.Val = t.Val;
    }
    break;
  case 'F':
    for (i = 0; i < sizeof(AppConfig.NetBIOSName); i++)
      Config.NetBIOSName[i] = AppConfig.NetBIOSName[i] = val[i];
    break;
  }
  ConfigSave();
}

void format_ip(char *str, int32 ip)
{
  IP_ADDR t;
  t.Val = ip;
  sprintf(str, "%u.%u.%u.%u", t.v[0], t.v[1], t.v[2], t.v[3]);
}

void tcpip_config_format_char(char id, char *str)
{
  switch (id)
  {
  case '0':
    sprintf(str, "%02X:%02X:%02X:%02X:%02X:%02X",
      MY_MAC_BYTE1, MY_MAC_BYTE2, MY_MAC_BYTE3,
      MY_MAC_BYTE4, MY_MAC_BYTE5, MY_MAC_BYTE6);
    break;
  case '1':
  case '2':
    if ((id == '2') == DHCPIsDisabled()) strcpy(str, " checked"); break;
  case '3': format_ip(str, AppConfig.MyIPAddr        .Val); break;
  case '4': format_ip(str, AppConfig.MyMask          .Val); break;
  case '5': format_ip(str, AppConfig.MyGateway       .Val); break;
  case '6': format_ip(str, AppConfig.PrimaryDNSServer.Val); break;
  case '7': memcpy(str, AppConfig.NetBIOSName, sizeof(AppConfig.NetBIOSName)); str[16] = 0; break;
  case '8': sprintf(str, "%u", ENCRevID); break;
  }
}

/*
 *  Time Configuration Page
 */

const char HTML_TIME_CONFIG_PAGE[]=
"%D<title>Time</title>%L%S>"
"<script type=\"text/javascript\">"
"function setnow(){"
"var e,f=document.forms[0],n=new Date();"
"for(e in f.elements){"
"if(e=='A')f[e].value=n.getFullYear();"
"if(e=='B')f[e].value=n.getMonth()+1;"
"if(e=='C')f[e].value=n.getDate();"
"if(e=='D')f[e].value=n.getHours();"
"if(e=='E')f[e].value=n.getMinutes();"
"if(e=='F')f[e].value=n.getSeconds();"
"if(e=='G')f[e].selectedIndex=24-n.getTimezoneOffset()/30;"
"}"
"f.submit();"
"};"
"onload=function(){"
"document.getElementById('set').innerHTML='<button onclick=\"setnow()\">Set to now<'+'/button>';"
"};"
"</script>"
"<h1>Time</h1>"
"<table border=1><tr><td><form action=\"/time\" method=post>"
"<p>Year:<input name=A size=4 value=\"%0\">"
"Month:<input name=B size=2 value=\"%1\">"
"Day:<input name=C size=2 value=\"%2\">"
"<p>Hour:<input name=D size=2 value=\"%3\">"
"Minute:<input name=E size=2 value=\"%4\">"
"Second:<input name=F size=2 value=\"%5\">"
"<p>Timezone:<select name=G>"
"%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6%6"
"</select>"
"<p><input type=submit value=Set>"
"</form></table>"
"<p id=set></p>"
"<p><a href=\"/\">Home</a>";

struct {
  int8 set;
  signed int16 cyr;
  signed int8 mo;
  signed int8 dy;
  signed int8 hr;
  signed int8 mn;
  signed int8 sc;
  signed int8 tz;
} tc;

void time_config_cgi(char *key, char *val)
{
  signed int16 v;
  v = a_to_l(val);
  switch (key[0])
  {
  case 'A': if (v >=   0 && v <= 9999) { tc.cyr = v; bit_set(tc.set, 0); } break;
  case 'B': if (v >=   1 && v <= 12  ) { tc.mo  = v; bit_set(tc.set, 1); } break;
  case 'C': if (v >=   1 && v <= 31  ) { tc.dy  = v; bit_set(tc.set, 2); } break;
  case 'D': if (v >=   0 && v <= 23  ) { tc.hr  = v; bit_set(tc.set, 3); } break;
  case 'E': if (v >=   0 && v <= 59  ) { tc.mn  = v; bit_set(tc.set, 4); } break;
  case 'F': if (v >=   0 && v <= 59  ) { tc.sc  = v; bit_set(tc.set, 5); } break;
  case 'G': if (v >= -48 && v <= 48  ) { tc.tz  = v; bit_set(tc.set, 6); } break;
  }
  if (tc.set == 0x7F && tc.dy <= maxDom(tc.cyr, tc.mo))
  {
    // localtime -> UTC
    applyTimezone(tc.cyr, tc.mo, tc.dy, tc.hr, tc.mn, tc.sc, -tc.tz);
    RtcSetTimezone(tc.tz);
    set_datetime(tc.cyr, tc.mo, tc.dy, tc.hr, tc.mn, tc.sc, 0);
    tc.set = 0;
  }
}

void time_config_format_char(char id, char *str)
{
  static signed int8 tz;
  static char sel[] = " selected";
  static char nosel[] = "";
  char *p;
  int16 i;
  signed int16 cyr;
  signed int8 mo, dy, hr, mn, sc;
  get_datetime(cyr, mo, dy, hr, mn, sc, i);
  // UTC -> localtime
  applyTimezone(cyr, mo, dy, hr, mn, sc, RtcGetTimezone());
  i = -1;
  switch (id)
  {
  case '0': i = cyr; break;
  case '1': i = mo ; break;
  case '2': i = dy ; break;
  case '3': i = hr ; break;
  case '4': i = mn ; break;
  case '5': i = sc ; tz = -48; break;
  case '6':
    sprintf(&str[40], "%03d:00", tz / 4);
    str[40] = (tz >= 0) ? '+' : '-';
    if (tz % 4 != 0) str[44] = '3';
    if (tz == RtcGetTimezone()) p = sel;
    else p = nosel;
    sprintf(str, "<option value=%03d%s>%s", tz, p, &str[40]);
    tz += 2;
    break;
  }
  if (i != -1) sprintf(str, "%lu", i);
  tc.set = 0;
}

/*
 *  Time Server Configuration Page
 */

const char HTML_TIMESERVER_CONFIG_PAGE[]=
"%D<title>Time Server</title>%L%S>"
"<h1>Time Server</h1>"
"<form action=\"/server\" method=post><table border=1>"
"<tr><th>Maximum clock difference to sync:<td><input name=A value=\"%1\">secs"
"<tr><th>Broadcast interval:<td><input name=B value=\"%2\">secs (0=disabled)"
"</table><p><input type=submit></form>%F";

void timeserver_config_cgi(char *key, char *val)
{
  int16 v;
  v = a_to_l(val);
  switch (key[0])
  {
  case 'A':
    Config.TimeServerGPS_MaxDiff = v;
    break;
  case 'B':
    Config.TimeServer_BroadcastInterval = v;
    break;
  }
  ConfigSave();
}

void timeserver_config_format_char(char id, char *str)
{
  int16 v;
  switch (id)
  {
  case '1': v = Config.TimeServerGPS_MaxDiff       ; break;
  case '2': v = Config.TimeServer_BroadcastInterval; break;
  }
  sprintf(str, "%lu", v);
}

/*
 * Login Configuration Page
 */

const char HTML_LOGIN_CONFIG_PAGE[]=
"%D<title>Login</title>%L%S>"
"<h1>Login</h1>"
"<form action=\"/login\" method=post><table border=1>"
"<tr><th>Username:<td><input name=A value=\"%0\" maxlength=9>"
"<tr><th>Password:<td><input type=password name=B value=\"%1\" maxlength=9>"
"</table><p><input type=submit></form>%F";

void login_config_cgi(char *key, char *val)
{
  char credentials[LOGIN_SIZE];
  char *p1;
  char *p2;
  val[LOGIN_SIZE / 2] = 0;
  memcpy(credentials, Config.Login, LOGIN_SIZE);
  p1 = strchr(credentials, ':');
  if (key[0] == 'A')
  {
    // username
    // move existing password to correct position for new username
    p2 = &credentials[strlen(val)];
    memmove(p2, p1, LOGIN_SIZE / 2 + 1);
    // copy new username into place
    strcpy(credentials, val);
    *p2 = ':';
  }
  if (key[0] == 'B')
  {
    // password
    // replace existing password with new one
    strcpy(p1 + 1, val);
  }
  // save new combination
  credentials[LOGIN_SIZE - 1] = 0;
  memcpy(Config.Login, credentials, LOGIN_SIZE);
  ConfigSave();
}

void login_config_format_char(char id, char *str)
{
  char credentials[LOGIN_SIZE];
  int8 i, j;
  char c;
  memcpy(credentials, Config.Login, LOGIN_SIZE);
  switch (id)
  {
  case '0':
    for (i = 0; i < LOGIN_SIZE; ++i)
    {
      c = credentials[i];
      if (c == ':' || c == 0) break;
      str[i] = c;
    }
    str[i] = 0;
    break;
  case '1':
    for (i = 0; i < LOGIN_SIZE; ++i)
    {
      c = credentials[i];
      if (c == ':' || c == 0) break;
    }
    for (++i, j = 0; i < LOGIN_SIZE; ++i, ++j)
    {
      c = credentials[i];
      if (c == 0) break;
      str[j] = c;
    }
    str[j] = 0;
    break;
  }
}

/*
 *  HTTP callbacks
 */

const char http_files[][*] = {
  "/","/status","/sats","/tcpip","/time","/server","/login","/styles"
};

void http_get_page(char *file_str, int32 *file, char *contentType)
{
  int8 file_idx, ch_idx;
  strcpy(contentType, "text/html;charset=utf-8");
  *file = 0;
  for (file_idx = 0; file_idx < 8; file_idx++)
  {
    for (ch_idx = 0; file_str[ch_idx] == http_files[file_idx][ch_idx]; ch_idx++)
    {
      if (file_str[ch_idx] == 0)
      {
        switch (file_idx)
        {
        case 0: *file = label_address(HTML_INDEX_PAGE); break;
        case 1: *file = label_address(HTML_STATUS_PAGE); break;
        case 2: *file = label_address(SVG_STATUS_PAGE); strcpy(contentType, "image/svg+xml"); break;
        case 3: *file = label_address(HTML_TCPIP_CONFIG_PAGE); break;
        case 4: *file = label_address(HTML_TIME_CONFIG_PAGE); break;
        case 5: *file = label_address(HTML_TIMESERVER_CONFIG_PAGE); break;
        case 6: *file = label_address(HTML_LOGIN_CONFIG_PAGE); break;
        case 7: *file = label_address(CSS_STYLE_SHEET); strcpy(contentType, "text/css"); break;
        }
        return;
      }
    }
  }
}

int1 http_check_authorization(int32 file, char *realm)
{
  if (realm != 0) strcpy(realm, "Administrator");
  else
  {
    char test[LOGIN_SIZE];
    // check for pages that are always available
    if (file == label_address(HTML_INDEX_PAGE) ||
        file == label_address(HTML_STATUS_PAGE) ||
        file == label_address(SVG_STATUS_PAGE) ||
        file == label_address(CSS_STYLE_SHEET)) return 1;
    // other pages may be restricted
    memcpy(test, Config.Login, LOGIN_SIZE);
    return (strcmp(test, http_auth_credentials) == 0);
  }
  return 0;
}

int1 http_exec_cgi(int32 file, char *key, char *val)
{
  int8 i = 0;
  if (file == label_address(HTML_TCPIP_CONFIG_PAGE     )) i = 1;
  if (file == label_address(HTML_TIME_CONFIG_PAGE      )) i = 2;
  if (file == label_address(HTML_TIMESERVER_CONFIG_PAGE)) i = 3;
  if (file == label_address(HTML_LOGIN_CONFIG_PAGE     )) i = 4;
  if (file == label_address(HTML_STATUS_PAGE           )) i = 5;
  if (file == label_address( SVG_STATUS_PAGE           )) i = 6;
  if (key == 0 || val == 0) return (i != 0);
  switch (i)
  {
  case 1:      tcpip_config_cgi(key, val); break;
  case 2:       time_config_cgi(key, val); break;
  case 3: timeserver_config_cgi(key, val); break;
  case 4:      login_config_cgi(key, val); break;
  }
  return 0;
}

int8 http_format_char(int32 file, char id, char *str, int8 max_ret)
{
  if (id == 'D')
  {
    strcpy(str, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\r\n");
  }
  else if (id == 'F')
  {
    strcpy(str, "<p><a href=\"/\">Home</a>");
    strcpy(&str[(file == label_address(HTML_INDEX_PAGE)) ? 0 : 23], "<p>GPS NTP Server v1.07"); // overwrite Home link on Index page
  }
  else if (id == 'L')
  {
    strcpy(str, "<link rel=stylesheet");
  }
  else if (id == 'S')
  {
    strcpy(str, " type=\"text/css\" href=\"/styles\"");
  }
  else
  {
    str[0] = 0;
    if (file == label_address(HTML_STATUS_PAGE           )) status_format_char(id, str);
    if (file == label_address( SVG_STATUS_PAGE           )) svg_status_format_char(id, str);
    if (file == label_address(HTML_TCPIP_CONFIG_PAGE     )) tcpip_config_format_char(id, str);
    if (file == label_address(HTML_TIME_CONFIG_PAGE      )) time_config_format_char(id, str);
    if (file == label_address(HTML_TIMESERVER_CONFIG_PAGE)) timeserver_config_format_char(id, str);
    if (file == label_address(HTML_LOGIN_CONFIG_PAGE     )) login_config_format_char(id, str);
  }
  return strlen(str);
}

/* * * * * * * * * * * * * * * *
 * Initialization & Main Loop  *
 * * * * * * * * * * * * * * * */

void init(void)
{
  int8 i;
  
  // make sure LEDs are set
  led_init();
  led_set(1, LED_RED);
  led_set(2, LED_OFF);
  
  // de-select all SPI devices
  output_high(PIN_EE1_CS);
  output_high(PIN_EE2_CS);
  output_high(PIN_EE3_CS);
  output_high(PIN_EE4_CS);
  output_high(PIN_ENC_MAC_CS);
  output_low(RTC_CE);
  output_drive(PIN_EE1_CS);
  output_drive(PIN_EE2_CS);
  output_drive(PIN_EE3_CS);
  output_drive(PIN_EE4_CS);
  output_drive(PIN_ENC_MAC_CS);
  output_drive(RTC_CE);
  exp_init();
  
  setup_timer_1(T1_INTERNAL | T1_DIV_BY_1);
  
  init_millis(); // setup millisecond tick
  pps_init();
  serial_init();
  output_drive(PIN_C3); // SPI clock
  output_drive(PIN_C5); // SPI data out
  setup_spi(SPI_MASTER|SPI_L_TO_H|SPI_XMIT_L_TO_H|SPI_CLK_DIV_4);
  
  // read the MAC address
  output_low(PIN_MAC_EE_CS);
  spi_write(0x03); // read
  spi_write(0xFA); // start of MAC address
  for (i = 0; i < 6; i++) AppConfig.MyMACAddr.v[i] = spi_read(0);
  output_high(PIN_MAC_EE_CS);
  ConfigLoad();
  AppConfig.MyIPAddr.Val         = Config.MyIPAddr;
  AppConfig.MyMask.Val           = Config.MyMask;
  AppConfig.MyGateway.Val        = Config.MyGateway;
  AppConfig.PrimaryDNSServer.Val = Config.PrimaryDNSServer;
  if (!Config.UseDHCP) DHCPDisable();
  memcpy(AppConfig.NetBIOSName, Config.NetBIOSName, sizeof(AppConfig.NetBIOSName));
  
  StackInit();
  #ifdef TELNET_PORT
  telnet_init();
  #endif
  TimeServerInit();
  SetCLKOUT(0x01); // 25MHz
  
  rtc_init();
  intrtc_init();
  write_ds1305(0x91, 0xA5); // trickle charge enable, one diode, 2K resistor
  i = read_ds1305(0x11);
  
  if (i == 0xA5 && ENCRevID == 4) status_ok = 1;
}

void main()
{
  init();
  for (;;)
  {
    intrtc_update();
    StackTask();
    #ifdef TELNET_PORT
    telnet_task();
    #endif
    TimeServerTask();
    if (pps_update_dac)
    {
      pps_update_dac = FALSE;
      output_low(PIN_DAC_CS);
      spi_write(make8(pps_dac_value, 1) | 0x10);
      spi_write(make8(pps_dac_value, 0));
      output_high(PIN_DAC_CS);
    }
  }
}
